#include "ipDCT2D.h"

namespace Torch {

inline real alphaDCT2D(int u, int N)
{
	if(u) return ((real) sqrt(2.0 / N));

	return ((real) sqrt(1.0 / N) );
}

void init_coeff_order(int block_size_, int *dct_)
{
	int x = 0, y = 0;
	int id = 0;

	dct_[2*id] = x;
	dct_[2*(id++)+1] = y;
	
	for(;;)
	{
		/* right or down ?  */
		if(x != block_size_-1) x = x + 1;
		else y = y + 1;
		
		dct_[2*id] = x;
		dct_[2*(id++)+1] = y;
		
		/* now down left..  */
		while(x != 0 && y != block_size_-1)
		{
			x = x - 1;
			y = y + 1;
			dct_[2*id] = x;
			dct_[2*(id++)+1] = y;
		}

		/* right or down ? */
		if(y == block_size_-1)
		{
			x = x + 1;
			dct_[2*id] = x;
			dct_[2*(id++)+1] = y;
			if(x == block_size_-1) break;
		}
		else
		{
			y = y + 1;
			dct_[2*id] = x;
			dct_[2*(id++)+1] = y;
		}

		/* finally up right..  */
		while(x != block_size_-1 && y != 0)
		{
			x = x + 1;
			y = y - 1;
			dct_[2*id] = x;
			dct_[2*(id++)+1] = y;
		}
	}
}

void ipDCT2D::init_()
{	
	// (x;y) coordinates of each DCT coefficient..
	dct_coeff_order = (int *)allocator->alloc(2 * block_size * block_size * sizeof(int));
	
	init_coeff_order(block_size, dct_coeff_order);
	
	// Init alpha and cosine table
	for(int u = 0; u < block_size; u++)
	{
	   	real alpha_u;

		alpha_u = alphaDCT2D(u, block_size);

		for(int v = 0; v < block_size; v++)
		{
	      		alpha[u][v] = alpha_u * alphaDCT2D(v, block_size);

			for(int x = 0; x < block_size; x++)
			{
			   	for(int y = 0; y < block_size; y++)  
				{
				   	// Note: we can pre-compute u*M_PI, v*M_PI and 2*N
					real tmpa = ( real(2*x+1) * (real)u * M_PI ) / real(2*block_size);
					real tmpb = ( real(2*y+1) * (real)v * M_PI ) / real(2*block_size);

					cosine[u][v][y][x] = cos(tmpa) * cos(tmpb);
				}
			}
		}
	}
}

ipDCT2D::ipDCT2D(int block_size_, const char *coding_, int dct_dim_) : ipCore(coding_)
{
	block_size = block_size_;

	if(block_size <  8)
		error("ipDCT2D(): Block size should not be smaller than 8.");

	if(block_size >  DCT2D_MAX_BLOCK)
		error("ipDCT2D(): Block size should not exceed %d.", DCT2D_MAX_BLOCK);

	dct_dim = dct_dim_;

	seq_out = new(allocator) Sequence(1, dct_dim);

	// init intern tab (alpha, cosine and dct_coeff_order)
	init_();

	if(coding_id != 0)
		warning("2D DCT not implemented for rgb, yuv and float coding yet.");
}

void ipDCT2D::process(Sequence *seq_in)
{
	real *dst = seq_out->frames[0];

	if(dct_dim > block_size * block_size)
		error("ipDCT2D(): dct_dim (%d) exceeds the total number of coefficients.", dct_dim);

	for(int i = 0; i < dct_dim; i++) 
	{
	   	int u = dct_coeff_order[2*i];
		int v = dct_coeff_order[2*i+1];
	
		real sum = 0.0;
		
		for(int y = 0; y < block_size; y++)  
		{
		   	int offset_ = y * block_size;
			
		   	for(int x = 0; x < block_size; x++)  
		   	{
			   	int idx_ = offset_ + x;
				
				sum += seq_in->frames[0][idx_] * cosine[u][v][x][y];
			}
		}
	
		*dst++ = alpha[u][v] * sum;
	}
}

ipDCT2D::~ipDCT2D()
{
}

}

