#ifndef IPDCT2D_INC
#define IPDCT2D_INC

#include "ipCore.h"

namespace Torch {

#define DCT2D_MAX_BLOCK 64

/** This class is designed to compute the 2D Discrete Cosine Transform

    This class computes 2D DCT over $n \times n$ images of size (where $n = 8 \dots 64$).
    
    \begin{verbatim}

      seq_in->frames[0]        +--------------+
	       +----+          |              |
	       |    |   ---->  |   ipDCT2D    | ----> seq_out->frames[0] (DCT coefficients)
	       |    |          |              |
	       +----+          +--------------+
             (8x8 image)
	 
    \end{verbatim}
    
    @author Sebastien Marcel (marcel@idiap.ch)
    @author Yann Rodriguez   (rodrig@idiap.ch)
    @author Conrad Sanderson (conradsand@ieee.org)
    
    @version 2.0
    \Date
    @since 2.0
*/
class ipDCT2D : public ipCore
{
   	// friend that can access private data
	friend class ipDCT2Dinverse;
	
	//
	void init_(void);

protected:
	// alpha coefficients
	real alpha[DCT2D_MAX_BLOCK][DCT2D_MAX_BLOCK];

	// cosine table
	real cosine[DCT2D_MAX_BLOCK][DCT2D_MAX_BLOCK][DCT2D_MAX_BLOCK][DCT2D_MAX_BLOCK];

	// DCT coefficient ordering
	int *dct_coeff_order;
public:
	
   	/// DCT block size
	int block_size; 
  
	/// number of DCT coefficients to keep
	int dct_dim;
	
	//----
	
	/** constructor
	
	    @param block_size_ is the input size of image to process
	    @param coding_ is the coding format of the image to process
            @param dct_dim_ is the number of DCT coefficient to keep
	*/
	ipDCT2D(int block_size_, const char *coding_, int dct_dim_);

	///
    	virtual ~ipDCT2D();

	///
    	virtual void process(Sequence *seq_in);

};

}

#endif
