#include "ipGaussian.h"
#include "ImageGray.h"


//*******************************************************************
//
// This file implements a gaussian filter (used with MSR)
//
// ---> author: Guillaume Heusch (guillaume.heusch@epfl.ch) <---
//
//*******************************************************************

namespace Torch {

ipGaussian::ipGaussian(int size_, int width_, int height_, const char *coding_) : ipCore(coding_)
{
  size = size_;
  width = width_;
  height = height_;

  halfsize = (int)ceil(size/2);
  sigma = (real)size/5;
  
  kernel = (real*) allocator->alloc (size*size*sizeof(real));

  real sum = 0.0;

  // build the original gaussian kernel
  for(int i = 0; i < size; i++) {
    
    real *kernel_line = kernel + (i*size);  // define a pointer on the kernel current line
    
    for (int j = 0; j < size; j++) {
      
      int x_value = (-halfsize) + i;
      int y_value = (-halfsize) + j;
      
      kernel_line[j] = exp(-(((real) (x_value * x_value + y_value * y_value)) / sigma));
      sum +=  kernel_line[j];
    }
  }

  for (int i=0; i<size*size; i++)
    kernel[i] /= sum;

  // allocate memory for output
  seq_out = new(allocator) Sequence(1, width*height);

  for(int i=0; i<seq_out->n_frames; i++)
     for(int j=0; j<seq_out->frame_size; j++)
       seq_out->frames[i][j] = 0.0; 
}


void ipGaussian::process(Sequence *seq_in)
{
 switch(coding_id)
    {
    case 0:
      {	
	real *src = seq_in->frames[0];
	
	// -------------------------------------------------------------------------------
	// CONVOLUTION
	// -------------------------------------------------------------------------------
	int k, kprime, l, lprime, index_k, index_l;
	int index_out = 0;
	real conv_sum = 0.0;

	for(k = 0; k < height; k++){
	  for(l = 0; l < width; l++){	
	    
	    for(kprime = -halfsize; kprime <= halfsize; kprime++){
	  
	      // mirror interpolation
	      index_k = k - kprime;	
	      if(index_k < 0)
		index_k = abs(index_k) - 1;
	      if(index_k >= height)
		index_k = (2 * height) - index_k - 1; 
	      
	  
	      for(lprime = -halfsize; lprime <= halfsize; lprime++){
		
		// mirror interpolation
		index_l = l-lprime;
		if(index_l < 0)
		  index_l = abs(index_l) - 1;
		if(index_l >= width)
		  index_l = (2 * width) - index_l - 1;
		
		real *image_line = src + (index_k*width);
		real *kernel_line = kernel + (kprime+halfsize)*size;
		
		conv_sum += (1.0/real(size*size))*kernel_line[(lprime+halfsize)]*image_line[index_l]; 
	      }
	    }
	    seq_out->frames[0][index_out]= conv_sum;
	    conv_sum = 0;
	    index_out++; 
	  } 
	}

	break;
      }
    case 1:
      message("ipGaussian::process() not implemented for RGB format");
      break;
    case 2:
      message("ipGaussian::process() not implemented for YUV format");
      break;
    case 3:
      message("ipGaussian::process() not implemented for float format");
      break;
    }
}

ipGaussian::~ipGaussian() 
{
}
}
