#include "ipRotate.h"

namespace Torch {

inline double ccwAngle_(double degAngle_)
{
   	double degAngle = degAngle_;
	
   	if(degAngle >= 360.0)
	{
	   	degAngle -= 360.0;
		while(degAngle >= 360.0) degAngle -= 360.0;
	}
	else if(degAngle < 0.0)
	{
	   	degAngle += 360.0;
    		while(degAngle < 0.0) degAngle += 360.0;
	}

	return degAngle;
}

ipRotate::ipRotate(real degAngle_, int width_, int height_, const char *coding_) : ipCore(coding_)
{
   	degAngle = degAngle_;

	width = width_;
	height = height_;

	width_out = -1;
	height_out = -1;
	
	init_();

	seq_out = new(allocator) Sequence(1, width_out*height_out*n_planes);
	
	for(int i = 0 ; i < seq_out->frame_size ; i++)
		seq_out->frames[0][i] = 0.0;
}

void ipRotate::init_(void)
{
	shift = new(allocator) ipShift(width, height, coding); 
	xc = FixI(width/2.0);
	yc = FixI(height/2.0);
	shiftx = 0;
	shifty = 0;
	
	pixmap1_ = NULL;
	pixmap2_ = NULL;
	pixmap3_ = NULL;

	clrBack.r = 0.0;
	clrBack.g = 0.0;
	clrBack.b = 0.0;
	
	// returns a Counter Clock Wise (CCW) angle
	degAngleOpposite = ccwAngle_(-degAngle);
	degAngle = ccwAngle_(degAngle);

	computeSinCosin(&rim, degAngle);
		
	// determines rotated output image size
	if(degAngle == 0.0)
	{
		width_out = width;
		height_out = height;
	}
	else if(degAngle == 180.0)
	{
		width_out = width;
		height_out = height;
	}
	else if(degAngle == 90.0)
	{
		width_out = height;
		height_out = width;
	}
	else if(degAngle == 270.0)
	{
		width_out = height;
		height_out = width;
	}
	else
	{
	   	double dAbsSin = fabs(rim.dSin);
		double dAbsCos = fabs(rim.dCos);
		
    		width_out = FixI((double)height * dAbsSin + (double)width * dAbsCos);
    		height_out = FixI((double)width * dAbsSin + (double)height * dAbsCos);
	}

	// decimate angles
	real angle_ = degAngleOpposite;

	if ((angle_ > 45.0) && (angle_ <= 135.0)) 
    	{
        	// Angle in (45.0 .. 135.0] 
	        // Rotate image by 90 degrees into temporary image,
        	// so it requires only an extra rotation angle 
	        // of -45.0 .. +45.0 to complete rotation.
    		
	   	width1_ = height;
    		height1_ = width;

		if(verbose) 
			print("preprocess (45.0 .. 135.0]\n");
		
	        angle_ -= 90.0;
    	}
    	else if ((angle_ > 135.0) && (angle_ <= 225.0)) 
    	{ 
        	// Angle in (135.0 .. 225.0] 
	        // Rotate image by 180 degrees into temporary image,
        	// so it requires only an extra rotation angle 
	        // of -45.0 .. +45.0 to complete rotation.
	   	
	   	width1_ = width;
    		height1_ = height;
	        
		if(verbose) 
			print("process (135.0 .. 225.0]\n");

		angle_ -= 180.0;
    	}
    	else if ((angle_ > 225.0) && (angle_ <= 315.0)) 
    	{	 
	        // Angle in (225.0 .. 315.0] 
        	// Rotate image by 270 degrees into temporary image,
	        // so it requires only an extra rotation angle 
        	// of -45.0 .. +45.0 to complete rotation.

	   	width1_ = height;
    		height1_ = width;
		
		if(verbose) 
			print("process (225.0 .. 315.0]\n");

        	angle_ -= 270.0;
    	}
	else
	{
	   	width1_ = width;
    		height1_ = height;
	}
	
	if(verbose) 
		print("reminding angle:%g\n", angle_);

	//
	computeSinCosin(&rsim, ccwAngle_(angle_));

	//
	pixmap1_ = (real *) allocator->alloc(n_planes * width1_ * height1_ * sizeof(real));
	
	//
    	width2_ = width1_ + FixI((double)height1_ * fabs(rsim.dTan2));
    	height2_ = height1_;
	pixmap2_ = (real *) allocator->alloc(n_planes * width2_ * height2_ * sizeof(real));

	//
    	width3_ = width2_;
    	height3_ = FixI((double)width1_ * fabs(rsim.dSin) + (double)height1_ * rsim.dCos);
	pixmap3_ = (real *) allocator->alloc(n_planes * width3_ * height3_ * sizeof(real));

	//
   	for(int i = 0 ; i < n_planes * width1_ * height1_ ; i++) pixmap1_[i] = 0.0;
   	for(int i = 0 ; i < n_planes * width2_ * height2_ ; i++) pixmap2_[i] = 0.0;
   	for(int i = 0 ; i < n_planes * width3_ * height3_ ; i++) pixmap3_[i] = 0.0;

	if(verbose) 
	{
		print("   width1_: %d\n", width1_);
		print("   height1_ : %d\n", height1_);
		print("   width2_: %d\n", width2_);
		print("   height2_ : %d\n", height2_);
		print("   width3_: %d\n", width3_);
		print("   height3_ : %d\n", height3_);
		print("   width_out: %d\n", width_out);
		print("   height_out : %d\n", height_out);
	}
   
	shiftx = FixI((width/2.0) - xc);
	shifty = FixI((height/2.0) - yc);
	
	shift->setIOption("shiftx", shiftx);
	shift->setIOption("shifty", shifty);

	if(verbose)
	{
		print("   (xc,yc) = (%d,%d)\n", xc, yc);
		print("   shiftx = %d\n", shiftx);
		print("   shifty = %d\n", shifty);
	}
	
	// Center of rotation of the src image
	rim.xc = xc;
  	rim.yc = yc;
		
	// Center of rotation of the dst image
   	rim.XC = (double) width_out / 2.0;
   	rim.YC = (double) height_out / 2.0;

	real x, y;
	int i;
	
	i = 0; x = 0; y = 0;
	rotate(x, y, &bbx[i].x, &bbx[i].y);
	bbx[i].x = FixI(bbx[i].x);
	bbx[i].y = FixI(bbx[i].y);
	if(verbose) print("bbx[%d]: (%g %g) -> (%g %g)\n", i, x, y, bbx[i].x, bbx[i].y);
	i++; x = width; y = 0;
	rotate(x, y, &bbx[i].x, &bbx[i].y);
	bbx[i].x = FixI(bbx[i].x);
	bbx[i].y = FixI(bbx[i].y);
	if(verbose) print("bbx[%d]: (%g %g) -> (%g %g)\n", i, x, y, bbx[i].x, bbx[i].y);
	i++; x = width; y = height;
	rotate(x, y, &bbx[i].x, &bbx[i].y);
	bbx[i].x = FixI(bbx[i].x);
	bbx[i].y = FixI(bbx[i].y);
	if(verbose) print("bbx[%d]: (%g %g) -> (%g %g)\n", i, x, y, bbx[i].x, bbx[i].y);
	i++; x = 0; y = height;
	rotate(x, y, &bbx[i].x, &bbx[i].y);
	bbx[i].x = FixI(bbx[i].x);
	bbx[i].y = FixI(bbx[i].y);
	if(verbose) print("bbx[%d]: (%g %g) -> (%g %g)\n", i, x, y, bbx[i].x, bbx[i].y);
	i++; bbx[i] = bbx[0];
}

void ipRotate::process(Sequence *seq_in)
{
	shift->process(seq_in);
	
	switch(coding_id)
	{
	case 0:
		rotateimage_gray(width, height, shift->seq_out->frames[0], width_out, height_out, seq_out->frames[0]);
		break;
		
	case 1:
		rotateimage_rgb(width, height, (rgbReal *) shift->seq_out->frames[0], width_out, height_out, (rgbReal *) seq_out->frames[0]);
		break;
		
	case 2:
		message("ipRotate::process() not implemented for YUV format.");
		break;
	case 3:
		message("ipRotate::process() not implemented for float format.");
		break;
	}
}

int ipRotate::getWidthOut(void)
{
        return width_out;
}

int ipRotate::getHeightOut(void)
{
        return height_out;
}

ipRotate::~ipRotate()
{
}

void ipRotate::HorizSkew_gray(real *pSrc, int width_src, int height_src, real *pDst, int width_dst, int height_dst, int uRow, int iOffset, double dWeight)
/*------------------------------------------------------------------------------

  Function: HorizSkew_gray

  Purpose:  Skews a row horizontally (with filtered weights)

  Input:    Image to skew (+ dimensions)
            Destination of skewed image (+ dimensions)
            Row index
            Skew offset
            Relative weight of right pixel
            Background color

  Output:   None.

  Remarks:  Limited to 45 degree skewing only. Filters two adjacent pixels.

------------------------------------------------------------------------------*/
{
   	int i;

	// Fill gap left of skew with background
    	for (i = 0; i < iOffset; i++)
		pDst[i + uRow * width_dst] = 0.0;
	   
    	real pxlOldLeft = 0.0;
    	for (i = 0; i < width_src ; i++) 
    	{
        	// Loop through row pixels
        	real pxlSrc;
		
		pxlSrc = pSrc[i + uRow * width_src];

        	// Calculate weights
        	real pxlLeft;

		pxlLeft = (real) (dWeight * (double) pxlSrc);
		
        	// Update left over on source
		pxlSrc = pxlSrc - (pxlLeft - pxlOldLeft);

        	// Check boundries 
        	if ((i + iOffset >= 0) && (i + iOffset < width_dst))
			pDst[i + iOffset + uRow * width_dst] = pxlSrc;
		   
        	// Save leftover for next pixel in scan
        	pxlOldLeft = pxlLeft;
    	}
	
    	// Go to rightmost point of skew
    	i += iOffset;  
	
        // If still in image bounds, put leftovers there
    	if (i < width_dst)
		pDst[i + uRow * width_dst] = pxlOldLeft;
	   
	// Clear to the right of the skewed line with background
    	while (++i < width_dst)
		pDst[i + uRow * width_dst] = 0.0;
	   
}   // HorizSkew_gray


void ipRotate::VertSkew_gray(real *pSrc, int width_src, int height_src, real *pDst, int width_dst, int height_dst, int uCol, int iOffset, double dWeight)
/*------------------------------------------------------------------------------

  Function: VertSkew_gray

  Purpose:  Skews a column vertically (with filtered weights)

  Input:    Image to skew (+dimensions)
            Destination of skewed image (+dimensions)
            Column index
            Skew offset
            Relative weight of upper pixel
            Background color

  Output:   None.

  Remarks:  Limited to 45 degree skewing only. Filters two adjacent pixels.

------------------------------------------------------------------------------*/
{
   	int i;
	
        // Fill gap above skew with background
    	for (i = 0; i < iOffset; i++)
		pDst[uCol + i * width_dst] = 0.0;
	   
    	real pxlOldLeft = 0.0;

    	int iYPos = 0;
    	for (i = 0; i < height_src; i++) 
    	{
        	// Loop through column pixels
        	real pxlSrc;
		
		pxlSrc = pSrc[uCol + i * width_src];
			
        	iYPos = i + iOffset;
		
        	// Calculate weights
        	real pxlLeft;
		
		pxlLeft = (real) (dWeight * (double) pxlSrc);
		
        	// Update left over on source
		pxlSrc = pxlSrc - (pxlLeft - pxlOldLeft);
		
        	// Check boundries
        	if ((iYPos >= 0) && (iYPos < height_dst))
			pDst[uCol + iYPos * width_dst] = pxlSrc;	
		   
        	// Save leftover for next pixel in scan
        	pxlOldLeft = pxlLeft;
    	}
	
    	// Go to bottom point of skew
    	i = iYPos;  
        
    	// If still in image bounds, put leftovers there
    	if (i < height_dst)
  		pDst[uCol + i * width_dst] = pxlOldLeft;
	   
        // Clear below skewed line with background
    	while (++i < height_dst)
  		pDst[uCol + i * width_dst] = 0.0;
	   
}   // VertSkew_gray

bool ipRotate::rotateimage0_gray(int width_src, int height_src, real *src_, int width_dst, int height_dst, real *dst_)
{
	int size = width_src * height_src;

   	while(size--) *dst_++ = *src_++;

	return true;
}

bool ipRotate::rotateimage90_gray(int width_src, int height_src, real *src_, int width_dst, int height_dst, real *dst_)
{
    	for (int uY = 0; uY < height_src ; uY++)
        	for (int uX = 0; uX < width_src ; uX++)
			dst_[uY + (height_dst - uX - 1) * width_dst] = src_[uX + uY * width_src];

	return true;
}

bool ipRotate::rotateimage180_gray(int width_src, int height_src, real *src_, int width_dst, int height_dst, real *dst_)
{
    	for (int uY = 0; uY < height_src ; uY++)
        	for (int uX = 0; uX < width_src ; uX++)
			dst_[(width_dst - uX - 1) + (height_dst - uY - 1) * width_dst] = src_[uX + uY * width_src];

	return true;
} 

bool ipRotate::rotateimage270_gray(int width_src, int height_src, real *src_, int width_dst, int height_dst, real *dst_)
{
    	for (int uY = 0; uY < height_src ; uY++)
        	for (int uX = 0; uX < width_src ; uX++)
			dst_[(width_dst - uY - 1) + uX * width_dst] = src_[uX + uY * width_src];
	
	return true;
} 

void ipRotate::rotateimage45_gray(real *pSrc, int width_src, int height_src, real *pDst, int width_dst, int height_dst, double dAngle)
/*------------------------------------------------------------------------------

  Function: rotateimage45_gray

  Purpose:  Rotates an image by a given degree in range [-45.0 .. +45.0]
            (counter clockwise)

  Remarks:  Using the 3-shear technique.

------------------------------------------------------------------------------*/
{
    	/******* Perform 1st shear (horizontal) ******/
	int u;

    	for (u = 0; u < height2_ ; u++) 
    	{  
        	double dShear;

        	if (rsim.dTan2 >= 0.0)
        	{
            		// Positive angle
            		dShear = ((double)u + 0.5) * rsim.dTan2;
        	}
        	else
        	{
            		// Negative angle
            		dShear = ((double) ((int)u - height2_) + 0.5) * rsim.dTan2;
        	}

        	int iShear = (int) floor(dShear);
		
        	HorizSkew_gray(	pSrc, width_src, height_src, 
		      		pixmap2_, width2_, height2_, 
				u, iShear, dShear - (double)iShear);
    	}

    
	/******* Perform 2nd shear  (vertical) ******/
    	double dOffset;     // Variable skew offset

    	if (rsim.dSin > 0.0)
    	{   
        	// Positive angle
        	dOffset = (double)(width_src - 1) * rsim.dSin;
    	}
    	else
    	{
        	// Negative angle
        	dOffset = -rsim.dSin * (double)(width_src - width3_);
    	}

    	for (u = 0; u < width3_ ; u++, dOffset -= rsim.dSin) 
    	{
        	int iShear = (int)floor(dOffset);
		
        	VertSkew_gray(	pixmap2_, width2_, height2_, 
		      		pixmap3_, width3_, height3_, 
				u, iShear, dOffset - (double)iShear);
    	}

	
    	/******* Perform 3rd shear (horizontal) ******/
    	if (rsim.dSin >= 0.0)
    	{
        	// Positive angle
        	dOffset = (double)(width_src - 1) * rsim.dSin * -rsim.dTan2;
    	}
    	else 
    	{
        	// Negative angle
        	dOffset = rsim.dTan2 * ((double)(width_src - 1) * -rsim.dSin + (double)(1 - height_dst));
    	}
	
    	for (u = 0; u < height_dst ; u++, dOffset += rsim.dTan2)
    	{
        	int iShear = (int)floor(dOffset);
		
        	HorizSkew_gray(	pixmap3_, width3_, height3_, 
    				pDst, width_dst, height_dst, 
		      		u, iShear, dOffset - double (iShear));
    	}
}   // rotateimage45_gray


bool ipRotate::rotateimage_gray(int width_src, int height_src, real *src_, int width_dst, int height_dst, real *dst_)
{
	real angle = degAngleOpposite;

	if(verbose)
		print("Go go rotate %g\n", angle);
	
	if(angle == 0.0)
	{
		if(verbose)
	   		print("process [0.0]\n");
		
		return rotateimage0_gray(width_src, height_src, src_, width_dst, height_dst, dst_);
	}
	
	if ((angle > 0.0) && (angle <= 45.0)) 
	{
		if(verbose)
	   		print("process (0.0 .. 45.0]\n");

		rotateimage45_gray(	src_, width_src, height_src, 
		      			dst_, width_dst, height_dst, 
					angle);

		if(verbose)
		{
			print("   width_out:%d\n", width_dst);
			print("   height_out:%d\n", height_dst);
		}

		return true;
	}
	
	if ((angle > 45.0) && (angle <= 135.0)) 
    	{
        	// Angle in (45.0 .. 135.0] 
	        // Rotate image by 90 degrees into temporary image,
        	// so it requires only an extra rotation angle 
	        // of -45.0 .. +45.0 to complete rotation.

		if(verbose)
	   		print("process (45.0 .. 135.0]\n");
	   
		rotateimage90_gray(width_src, height_src, src_, width1_, height1_, pixmap1_);
		
	        angle -= 90.0;
    	}
    	else if ((angle > 135.0) && (angle <= 225.0)) 
    	{ 
        	// Angle in (135.0 .. 225.0] 
	        // Rotate image by 180 degrees into temporary image,
        	// so it requires only an extra rotation angle 
	        // of -45.0 .. +45.0 to complete rotation.
		
		if(verbose)
	   		print("process (135.0 .. 225.0]\n");
        	
		rotateimage180_gray(width_src, height_src, src_, width1_, height1_, pixmap1_);

	        angle -= 180.0;
    	}
    	else if ((angle > 225.0) && (angle <= 315.0)) 
    	{	 
	        // Angle in (225.0 .. 315.0] 
        	// Rotate image by 270 degrees into temporary image,
	        // so it requires only an extra rotation angle 
        	// of -45.0 .. +45.0 to complete rotation.
		
		if(verbose)
	   		print("process (225.0 .. 315.0]\n");
	        
		rotateimage270_gray(width_src, height_src, src_, width1_, height1_, pixmap1_);
		
        	angle -= 270.0;
    	}
      	else if( angle > 315.0 )
      	{
		if(verbose)
	   		print("process (315.0 .. 360.0]\n");
		
        	angle -= 360.0;
		
		rotateimage45_gray(	src_, width_src, height_src, 
		      			dst_, width_dst, height_dst, 
					angle);

		if(verbose)
		{
			print("   width_out: %d\n", width_dst);
			print("   height_out: %d\n", height_dst);
		}

		return true;
	}
	
	if(verbose)
		print("   reminding angle: %g\n", angle);
    	
	if (angle == 0.0)
    	{
        	// No rotation at all
		return rotateimage0_gray(width1_, height1_, pixmap1_, width_dst, height_dst, dst_);
    	}
	else
	{
		rotateimage45_gray(	pixmap1_, width1_, height1_, 
		      			dst_, width_dst, height_dst, 
					angle);

		if(verbose)
		{
			print("   width_out: %d\n", width_dst);
			print("   height_out: %d\n", height_dst);
		}

		return true;
	}
}

void ipRotate::HorizSkew_rgb(rgbReal *pSrc, int width_src, int height_src, rgbReal *pDst, int width_dst, int height_dst, int uRow, int iOffset, double dWeight)
/*------------------------------------------------------------------------------

  Function: HorizSkew_rgb

  Purpose:  Skews a row horizontally (with filtered weights)

  Input:    Image to skew (+ dimensions)
            Destination of skewed image (+ dimensions)
            Row index
            Skew offset
            Relative weight of right pixel
            Background color

  Output:   None.

  Remarks:  Limited to 45 degree skewing only. Filters two adjacent pixels.

------------------------------------------------------------------------------*/
{
   	int i;

	// Fill gap left of skew with background
    	for (i = 0; i < iOffset; i++)
		pDst[i + uRow * width_dst] = clrBack;

    	rgbReal pxlOldLeft = clrBack;
    	for (i = 0; i < width_src ; i++) 
    	{
        	// Loop through row pixels
        	rgbReal pxlSrc;
		
		pxlSrc = pSrc[i + uRow * width_src];

        	// Calculate weights
        	rgbReal pxlLeft;

		pxlLeft.r = (real) (dWeight * (double) pxlSrc.r);
		pxlLeft.g = (real) (dWeight * (double) pxlSrc.g);
		pxlLeft.b = (real) (dWeight * (double) pxlSrc.b);
		
        	// Update left over on source
		pxlSrc.r = pxlSrc.r - (pxlLeft.r - pxlOldLeft.r);
		pxlSrc.g = pxlSrc.g - (pxlLeft.g - pxlOldLeft.g);
		pxlSrc.b = pxlSrc.b - (pxlLeft.b - pxlOldLeft.b);

        	// Check boundries 
        	if ((i + iOffset >= 0) && (i + iOffset < width_dst))
			pDst[i + iOffset + uRow * width_dst] = pxlSrc;
		   
        	// Save leftover for next pixel in scan
        	pxlOldLeft = pxlLeft;
    	}
	
    	// Go to rightmost point of skew
    	i += iOffset;  
	
        // If still in image bounds, put leftovers there
    	if (i < width_dst)
		pDst[i + uRow * width_dst] = pxlOldLeft;
	   
	// Clear to the right of the skewed line with background
    	while (++i < width_dst)
		pDst[i + uRow * width_dst] = clrBack;
	   
}   // HorizSkew_gray


void ipRotate::VertSkew_rgb(rgbReal *pSrc, int width_src, int height_src, rgbReal *pDst, int width_dst, int height_dst, int uCol, int iOffset, double dWeight)
/*------------------------------------------------------------------------------

  Function: VertSkew_rgb

  Purpose:  Skews a column vertically (with filtered weights)

  Input:    Image to skew (+dimensions)
            Destination of skewed image (+dimensions)
            Column index
            Skew offset
            Relative weight of upper pixel
            Background color

  Output:   None.

  Remarks:  Limited to 45 degree skewing only. Filters two adjacent pixels.

------------------------------------------------------------------------------*/
{
   	int i;
	
        // Fill gap above skew with background
    	for (i = 0; i < iOffset; i++)
		pDst[uCol + i * width_dst] = clrBack;
	   
    	rgbReal pxlOldLeft = clrBack;

    	int iYPos = 0;
    	for (i = 0; i < height_src; i++) 
    	{
        	// Loop through column pixels
        	rgbReal pxlSrc;
		
		pxlSrc = pSrc[uCol + i * width_src];
			
        	iYPos = i + iOffset;
		
        	// Calculate weights
        	rgbReal pxlLeft;
		
		pxlLeft.r = (real) (dWeight * (double) pxlSrc.r);
		pxlLeft.g = (real) (dWeight * (double) pxlSrc.g);
		pxlLeft.b = (real) (dWeight * (double) pxlSrc.b);
		
        	// Update left over on source
		pxlSrc.r = pxlSrc.r - (pxlLeft.r - pxlOldLeft.r);
		pxlSrc.g = pxlSrc.g - (pxlLeft.g - pxlOldLeft.g);
		pxlSrc.b = pxlSrc.b - (pxlLeft.b - pxlOldLeft.b);
		
        	// Check boundries
        	if ((iYPos >= 0) && (iYPos < height_dst))
			pDst[uCol + iYPos * width_dst] = pxlSrc;	
		   
        	// Save leftover for next pixel in scan
        	pxlOldLeft = pxlLeft;
    	}
	
    	// Go to bottom point of skew
    	i = iYPos;  
        
    	// If still in image bounds, put leftovers there
    	if (i < height_dst)
  		pDst[uCol + i * width_dst] = pxlOldLeft;
	   
        // Clear below skewed line with background
    	while (++i < height_dst)
  		pDst[uCol + i * width_dst] = clrBack;
	   
}   // VertSkew_gray

bool ipRotate::rotateimage0_rgb(int width_src, int height_src, rgbReal *src_, int width_dst, int height_dst, rgbReal *dst_)
{
	int size = width_src * height_src;

   	while(size--) *dst_++ = *src_++;

	return true;
}

bool ipRotate::rotateimage90_rgb(int width_src, int height_src, rgbReal *src_, int width_dst, int height_dst, rgbReal *dst_)
{
    	for (int uY = 0; uY < height_src ; uY++)
        	for (int uX = 0; uX < width_src ; uX++)
			dst_[uY + (height_dst - uX - 1) * width_dst] = src_[uX + uY * width_src];

	return true;
}

bool ipRotate::rotateimage180_rgb(int width_src, int height_src, rgbReal *src_, int width_dst, int height_dst, rgbReal *dst_)
{
    	for (int uY = 0; uY < height_src ; uY++)
        	for (int uX = 0; uX < width_src ; uX++)
			dst_[(width_dst - uX - 1) + (height_dst - uY - 1) * width_dst] = src_[uX + uY * width_src];

	return true;
} 

bool ipRotate::rotateimage270_rgb(int width_src, int height_src, rgbReal *src_, int width_dst, int height_dst, rgbReal *dst_)
{
    	for (int uY = 0; uY < height_src ; uY++)
        	for (int uX = 0; uX < width_src ; uX++)
			dst_[(width_dst - uY - 1) + uX * width_dst] = src_[uX + uY * width_src];
	
	return true;
} 

void ipRotate::rotateimage45_rgb(rgbReal *pSrc, int width_src, int height_src, rgbReal *pDst, int width_dst, int height_dst, double dAngle)
/*------------------------------------------------------------------------------

  Function: rotateimage45_rgb

  Purpose:  Rotates an image by a given degree in range [-45.0 .. +45.0]
            (counter clockwise)

  Remarks:  Using the 3-shear technique.

------------------------------------------------------------------------------*/
{
    	/******* Perform 1st shear (horizontal) ******/
	int u;

    	for (u = 0; u < height2_ ; u++) 
    	{  
        	double dShear;

        	if (rsim.dTan2 >= 0.0)
        	{
            		// Positive angle
            		dShear = ((double)u + 0.5) * rsim.dTan2;
        	}
        	else
        	{
            		// Negative angle
            		dShear = ((double) ((int)u - height2_) + 0.5) * rsim.dTan2;
        	}

        	int iShear = (int) floor(dShear);
		
        	HorizSkew_rgb(	pSrc, width_src, height_src, 
		      		(rgbReal *) pixmap2_, width2_, height2_, 
				u, iShear, dShear - (double)iShear);
    	}

    
	/******* Perform 2nd shear  (vertical) ******/
    	double dOffset;     // Variable skew offset

    	if (rsim.dSin > 0.0)
    	{   
        	// Positive angle
        	dOffset = (double)(width_src - 1) * rsim.dSin;
    	}
    	else
    	{
        	// Negative angle
        	dOffset = -rsim.dSin * (double)(width_src - width3_);
    	}

    	for (u = 0; u < width3_ ; u++, dOffset -= rsim.dSin) 
    	{
        	int iShear = (int)floor(dOffset);
		
        	VertSkew_rgb(	(rgbReal *) pixmap2_, width2_, height2_, 
		      		(rgbReal *) pixmap3_, width3_, height3_, 
				u, iShear, dOffset - (double)iShear);
    	}

	
    	/******* Perform 3rd shear (horizontal) ******/
    	if (rsim.dSin >= 0.0)
    	{
        	// Positive angle
        	dOffset = (double)(width_src - 1) * rsim.dSin * -rsim.dTan2;
    	}
    	else 
    	{
        	// Negative angle
        	dOffset = rsim.dTan2 * ((double)(width_src - 1) * -rsim.dSin + (double)(1 - height_dst));
    	}
	
    	for (u = 0; u < height_dst ; u++, dOffset += rsim.dTan2)
    	{
        	int iShear = (int)floor(dOffset);
		
        	HorizSkew_rgb(	(rgbReal *) pixmap3_, width3_, height3_, 
    				pDst, width_dst, height_dst, 
		      		u, iShear, dOffset - double (iShear));
    	}
}   // rotateimage45_gray


bool ipRotate::rotateimage_rgb(int width_src, int height_src, rgbReal *src_, int width_dst, int height_dst, rgbReal *dst_)
{
	real angle = degAngleOpposite;

	if(verbose)
		print("Go go rotate %g\n", angle);
	
	if(angle == 0.0)
	{
		if(verbose)
	   		print("process [0.0]\n");
		
		return rotateimage0_rgb(width_src, height_src, src_, width_dst, height_dst, dst_);
	}
	
	if ((angle > 0.0) && (angle <= 45.0)) 
	{
		if(verbose)
	   		print("process (0.0 .. 45.0]\n");

		rotateimage45_rgb(	src_, width_src, height_src, 
		      			dst_, width_dst, height_dst, 
					angle);

		if(verbose)
		{
			print("   width_out:%d\n", width_dst);
			print("   height_out:%d\n", height_dst);
		}

		return true;
	}
	
	if ((angle > 45.0) && (angle <= 135.0)) 
    	{
        	// Angle in (45.0 .. 135.0] 
	        // Rotate image by 90 degrees into temporary image,
        	// so it requires only an extra rotation angle 
	        // of -45.0 .. +45.0 to complete rotation.

		if(verbose)
	   		print("process (45.0 .. 135.0]\n");
	   
		rotateimage90_rgb(width_src, height_src, src_, width1_, height1_, (rgbReal *) pixmap1_);
		
	        angle -= 90.0;
    	}
    	else if ((angle > 135.0) && (angle <= 225.0)) 
    	{ 
        	// Angle in (135.0 .. 225.0] 
	        // Rotate image by 180 degrees into temporary image,
        	// so it requires only an extra rotation angle 
	        // of -45.0 .. +45.0 to complete rotation.
		
		if(verbose)
	   		print("process (135.0 .. 225.0]\n");
        	
		rotateimage180_rgb(width_src, height_src, src_, width1_, height1_, (rgbReal *) pixmap1_);

	        angle -= 180.0;
    	}
    	else if ((angle > 225.0) && (angle <= 315.0)) 
    	{	 
	        // Angle in (225.0 .. 315.0] 
        	// Rotate image by 270 degrees into temporary image,
	        // so it requires only an extra rotation angle 
        	// of -45.0 .. +45.0 to complete rotation.
		
		if(verbose)
	   		print("process (225.0 .. 315.0]\n");
	        
		rotateimage270_rgb(width_src, height_src, src_, width1_, height1_, (rgbReal *) pixmap1_);
		
        	angle -= 270.0;
    	}
      	else if( angle > 315.0 )
      	{
		if(verbose)
	   		print("process (315.0 .. 360.0]\n");
		
        	angle -= 360.0;
		
		rotateimage45_rgb(	src_, width_src, height_src, 
		      			dst_, width_dst, height_dst, 
					angle);

		if(verbose)
		{
			print("   width_out: %d\n", width_dst);
			print("   height_out: %d\n", height_dst);
		}

		return true;
	}
	
	if(verbose)
		print("   reminding angle: %g\n", angle);
    	
	if (angle == 0.0)
    	{
        	// No rotation at all
		return rotateimage0_rgb(width1_, height1_, (rgbReal *) pixmap1_, width_dst, height_dst, dst_);
    	}
	else
	{
		rotateimage45_rgb(	(rgbReal *) pixmap1_, width1_, height1_, 
		      			dst_, width_dst, height_dst, 
					angle);

		if(verbose)
		{
			print("   width_out: %d\n", width_dst);
			print("   height_out: %d\n", height_dst);
		}

		return true;
	}
}

void ipRotate::computeSinCosin(ipRotateMatrix2D *r, double d_angle)
{
   	r->degAngle = d_angle;
	
	if((int) r->degAngle == 0)
	{
		r->radAngle = 0.0;
		r->dSin = 0.0;
		r->dCos = 1.0;
		r->dSinOpposite = 0.0;
		r->dCosOpposite = 1.0;
	}
	else if(r->degAngle == 180.0)
	{
		r->radAngle = M_PI;
		r->dSin = 0.0;
		r->dCos = -1.0;
		r->dSinOpposite = 0.0;
		r->dCosOpposite = -1.0;
	}
	else if(r->degAngle == 90.0)
	{
		r->radAngle = M_PI_2;
		r->dSin = 1.0;
		r->dCos = 0.0;
		r->dSinOpposite = -1.0;
		r->dCosOpposite = 0.0;
	}
	else if(r->degAngle == 270.0)
	{
		r->radAngle = 3.0 * M_PI_2;
		r->dSin = -1.0;
		r->dCos = 0.0;
		r->dSinOpposite = 1.0;
		r->dCosOpposite = 0.0;
	}
	else
	{
		// convert degrees to radians
		r->radAngle = r->degAngle * M_PI / 180.0;
		r->dSin = sin(r->radAngle);
		r->dCos = cos(r->radAngle);
		r->dSinOpposite = sin(-r->radAngle);
		r->dCosOpposite = cos(-r->radAngle);
	}

	r->dTan2 = tan(r->radAngle/2.0);

	r->xc = r->yc = 0.0;
	r->XC = r->YC = 0.0;
}

void ipRotate::rotate(int x, int y, int *X, int *Y)
{
	double xo, yo;
	double xr, yr;

   	// Changing coordinate system from src image to unity circle
   	xo = (double) x - rim.xc;
	yo = (double) y - rim.yc;
			
	// Performing rotation
	xr = rim.dCos * xo - rim.dSin * yo;
	yr = rim.dSin * xo + rim.dCos * yo;
	
   	// Changing coordinate system from unity circle to dest image 
  	*X = (int) (xr + rim.XC);
	*Y = (int) (yr + rim.YC);
}

void ipRotate::rotate(double x, double y, double *X, double *Y)
{
	double xo, yo;
	double xr, yr;

   	// Changing coordinate system from src image to unity circle
   	xo = x - rim.xc;
	yo = y - rim.yc;
			
	// Performing rotation
	xr = rim.dCos * xo - rim.dSin * yo;
	yr = rim.dSin * xo + rim.dCos * yo;
	
   	// Changing coordinate system from unity circle to dest image 
  	*X = xr + rim.XC;
	*Y = yr + rim.YC;
}

void ipRotate::unrotate(int x, int y, int *X, int *Y)
{
	double xo, yo;
	double xr, yr;

   	// Changing coordinate system from dst image to unity circle
   	xo = (double) x - rim.XC;
	yo = (double) y - rim.YC;
			
	// Performing rotation
	xr = rim.dCosOpposite * xo - rim.dSinOpposite * yo;
	yr = rim.dSinOpposite * xo + rim.dCosOpposite * yo;
	
   	// Changing coordinate system from unity circle to src image 
  	*X = (int) (xr + rim.xc);
	*Y = (int) (yr + rim.yc);
}

void ipRotate::unrotate(double x, double y, double *X, double *Y)
{
	double xo, yo;
	double xr, yr;

   	// Changing coordinate system from dst image to unity circle
   	xo = x - rim.XC;
	yo = y - rim.YC;
			
	// Performing rotation
	xr = rim.dCosOpposite * xo - rim.dSinOpposite * yo;
	yr = rim.dSinOpposite * xo + rim.dCosOpposite * yo;
	
   	// Changing coordinate system from unity circle to src image 
  	*X = xr + rim.xc;
	*Y = yr + rim.yc;
}


}

