#include "MultiScaleWindowScan.h"

namespace Torch {

#define IISUM_(P1, P2, P3, P4) (*(P4) - *(P3) - *(P2) + *(P1))
	
MultiScaleWindowScan::MultiScaleWindowScan(
		int width_image_, int height_image_, 
		int real_width_image_, int real_height_image_, 
		int width_pattern_, int height_pattern_, 
		int real_width_pattern_, int real_height_pattern_,
		real stepx_factor_, real stepy_factor_,
		ipSubWindow *ip_sw_, 
		PatternCollector *patterncollector_,
		bool verbose)
		: WindowScan(width_image_, height_image_, real_width_image_, real_height_image_, width_pattern_, height_pattern_, real_width_pattern_, real_height_pattern_, stepx_factor_, stepy_factor_, ip_sw_, patterncollector_, verbose)
{
	if(verbose)
	{
		print("MultiScaleWindowScan::MultiScaleWindowScan() scan %dx%d image (real=%dx%d) with %dx%d window (real=%dx%d)\n", 
				width_image, height_image, 
				real_width_image, real_height_image, 
				width_pattern, height_pattern,
				real_width_pattern, real_height_pattern);
		print(" + stepx = %d\n", stepx);
                print(" + stepy = %d\n", stepy);
	}
}

void MultiScaleWindowScan::init(void)
{
	ip_sw->init();
}

void MultiScaleWindowScan::process(Sequence *seq_in)
{
	// set the image size
	ip_sw->width_image = width_image;
	ip_sw->height_image = height_image;
	
	// set the window size 
	ip_sw->w_window = width_pattern;
	ip_sw->h_window = height_pattern;
	
	int scan_width = width_image - width_pattern;
	int scan_height = height_image - height_pattern;

	//
	int y;
	int x;

	//
	n_subwindows = 0;
	n_subwindows_processed = 0;
	n_pruning_reject = 0;
        n_mean_reject = 0;
        n_stdv_reject = 0;
	
	y = 2;

	// speed-up
	int width_pattern_1 = width_pattern-1;
	int height_pattern_1 = height_pattern-1;
	real *pruning_ii_p1=NULL, *pruning_ii_p2=NULL;
        real *pruning_ii_p3=NULL, *pruning_ii_p4=NULL;
	real *ii_p1, *ii_p2, *ii_p3, *ii_p4;
	real *ii2_p1, *ii2_p2, *ii2_p3, *ii2_p4;

	while(y < scan_height-1)
	{
	    int offset_y_width = (y-1) * real_width_image;
	    int offset_y_h_width = (y + height_pattern_1) * real_width_image;

	    if(use_pruning)
	    {
		pruning_ii_p1 = pruning_ii_p2 = pruning_ii + offset_y_width;
		pruning_ii_p3 = pruning_ii_p4 = pruning_ii + offset_y_h_width;
	    }

	    ii_p1 = ii_p2 = integralimage + offset_y_width;
	    ii_p3 = ii_p4 = integralimage + offset_y_h_width;
	    ii2_p1 = ii2_p2 = integralimage2 + offset_y_width;
	    ii2_p3 = ii2_p4 = integralimage2 + offset_y_h_width;

   	    x = 2;

	    while(x < scan_width-1)
	    {
		n_subwindows++;
		    
 		// speed-up
		int x_1 = x - 1;
		int offset_x_w = x + width_pattern_1;
		
		// PRUNING FILTER
		if(use_pruning)
		{
			real pruning_sum = IISUM_(pruning_ii_p1 + x_1, pruning_ii_p2 + offset_x_w, pruning_ii_p3 + x_1, pruning_ii_p4 + offset_x_w);

			if((pruning_sum < pruning_min_threshold) || (pruning_sum > pruning_max_threshold))
			{
				x += stepx;
				n_pruning_reject++;
				continue;
			}
		}

		// MEAN-STDV FILTER
		if(use_mean_stdv)
		{	
		    real mean = IISUM_(ii_p1 + x_1, ii_p2 + offset_x_w, ii_p3 + x_1, ii_p4 + offset_x_w);

		    if((mean < minMean) || (mean > maxMean))
		    {
			    x += stepx;
			    n_mean_reject++;
			    continue;
		    }
		    
		    real stdv = IISUM_(ii2_p1 + x_1, ii2_p2 + offset_x_w, ii2_p3 + x_1, ii2_p4 + offset_x_w);

		    stdv -= mean*mean*one_by_size_pattern;

		    if((stdv < minStdv) || (stdv > maxStdv))
		    {
			    x += stepx;
			    n_stdv_reject++;
			    continue;
		    }
		}

		// AND FINALLY THE MACHINE 
		// set the window location
    	        ip_sw->x_window = x;
		ip_sw->y_window = y;

		// processing subwindow
		ip_sw->process(seq_in);

		//if(ip_sw->seq_out->frames[0][0] > ip_sw->threshold)
		if(ip_sw->isPattern())
		{
			// add a new pattern
			if(patterncollector)
		            patterncollector->add(x, y, real_width_pattern, real_height_pattern, ip_sw->seq_out->frames[0][0], ip_sw->view);
		}
		
    		n_subwindows_processed++;
		
		x += stepx;
	    }
	    y += stepy;
	}
}

MultiScaleWindowScan::~MultiScaleWindowScan()
{
}

}

