#ifndef IP_HAAR_INC
#define IP_HAAR_INC

#include "ipCore.h"

namespace Torch {

/** This class is designed to compute one Haar-like feature into an integral image

    It seems long and tedious, but implemented to do a lot of precomputations and then speed up the execution.

    \begin{verbatim}

         pattern image	
		               #width_pattern#
	           +--------------------------------------+
	           |                                      |
	           |                                      |
	           |     #x_mask#                         |
	           | #y_mask# +--------------+            |
		   |          |              |            |
		   |          |              |            |
		   |          |              | #h_mask#   |
		   |          |              |            |
		   |          |              |            |
 #height_pattern#  |          |              |            |
	           |          +--------------+            |
	           |              #w_mask#                |
	           |                                      |
	           |                                      |
	           .                                      .
	           .                                      .
	           .                                      .
	           .                                      .
	           .                                      .
	           .                                      .
	           |                                      |
	           |                                      |
	           +--------------------------------------+



         main image	
				          #width_image# (>= #width_pattern#)
	                  +-----------------------------------------------------------+
	                  |                                                           |
	                  |                                                           |
	                  |                                                           |
	                  |    #x_window#                                             |
	                  |   #y_window# +-------------+                              |
	                  |              |             |                              |
	                  |              |    +---+    |                              |
    #height_image#	  |              |    |X| |    | #h_window#                   |
 (>= #height_pattern#)	  |              |    |X| |    |                              |
	                  |              |    +-+-+    |                              |
	                  |              |             |                              |
	                  |              +-------------+                              |
	                  |                 #w_window#                                |
	                  .                                                           .
	                  .                                                           .
	                  .                                                           .
	                  .                                                           .
	                  .                                                           .
	                  .                                                           .
	                  .                                                           .
	                  .                                                           .
	                  |                                                           |
	                  |                                                           |
	                  +-----------------------------------------------------------+


	Bi-Vertical, Bi-Horizontal, and Diagonal masks can be computed using this representation:
	                  
 			P1      P2      P3
			  o       o       o           
	                   +-------+-------+
		           |       |       |
		        P4 |    P5 |    P6 |
		          o|      o|      o|
		           +-------+-------+
		           |       |       |
		        P7 |    P8 |    P9 |
		          o|      o|      o|
	                   +-------+-------+


	Bi-Vertical mask:
	                   +-------+-------+
		           |XXXXXXX|       |
		           |XXXXXXX|       |
		           |XXXXXXX|       |
		           |XXXXXXX|       |
		           |XXXXXXX|       |
		           |XXXXXXX|       |
	                   +-------+-------+

	Bi-Horizontal mask:
	                   +---------------+
		           |XXXXXXXXXXXXXXX|
		           |XXXXXXXXXXXXXXX|
		           |XXXXXXXXXXXXXXX|
		           +---------------+
		           |               |
		           |               |
		           |               |
	                   +---------------+

	Diagonal mask:
	                   +-------+-------+
		           |XXXXXXX|       |
		           |XXXXXXX|       |
		           |XXXXXXX|       |
		           +-------+-------+
		           |       |XXXXXXX|
		           |       |XXXXXXX|
		           |       |XXXXXXX|
	                   +-------+-------+

	Tri-Vertical and Tri-Horizontal masks can be computed using this representation:
	                  
 			 P1   P12    P23    P3
			  o     o      o     o           
	                   +-----+-----+-----+
		           |     |     |     |
		       P14 |     |     | P36 |
		          o|     |     |    o|
	                   +-----+-----+-----+
		           |     |     |     |
		       P47 |     |     | P69 |
		          o|     |     |    o|
	                   +-----+-----+-----+
		           |     |     |     |
		        P7 |  P78|  P89|  P9 |
		          o|    o|    o|    o|
	                   +-----+-----+-----+

	Tri-Vertical mask:
	                   +-----+-----+-----+
		           |     |XXXXX|     |
		           |     |XXXXX|     |
		           |     |XXXXX|     |
		           |     |XXXXX|     |
		           |     |XXXXX|     |
		           |     |XXXXX|     |
		           |     |XXXXX|     |
		           |     |XXXXX|     |
		           |     |XXXXX|     |
		           |     |XXXXX|     |
		           |     |XXXXX|     |
	                   +-----+-----+-----+

	Tri-Horizontal mask:
	                   +-----------------+
		           |                 |
		           |                 |
		           |                 |
	                   +-----------------+
		           |XXXXXXXXXXXXXXXXX|
		           |XXXXXXXXXXXXXXXXX|
		           |XXXXXXXXXXXXXXXXX|
	                   +-----------------+
		           |                 |
		           |                 |
		           |                 |
	                   +-----------------+

	Central mask:
	                   +-----------------+
		           |                 |
		           |                 |
	                   |    +-------+    |
		           |    |XXXXXXX|    |
		           |    |XXXXXXX|    |
		           |    |XXXXXXX|    |
	                   |    +-------+    |
		           |                 |
		           |                 |
	                   +-----------------+
    \end{verbatim}

    @author Yann Rodriguez (yann.rodriguez@idiap.ch)
    @author Sebastien Marcel (marcel@idiap.ch)
    @version 2.0
    \Date
    @since 2.0
*/
class ipHaar : public ipCore
{
	// intern tmp variables for speed-up
	int x_mask_1;
	int y_mask_1;

	//
	real scale_x;
	real scale_y;
	real scale_xy_inv;

	//
	int x, y, w, h;

	//
	int w2, h2;
	int w3, h3;
	int w3_2, h3_2;
	
public:
	/// width of the image to scan
	int width_image;

	/// height of the image to scan
        int height_image;

	/// width of the model
	int width_pattern;

	/// height of the model
        int height_pattern;
	
	/// mask type
	int mask_id;

	/** @name location and size of the mask in the scanning window */
	//@{
	///
	int x_mask;

	///
	int y_mask;

	///
	int w_mask;

	///
	int h_mask;
	//@}
	
	/** @name location and size of the scanning window in the image */
	//@{
	///
	int x_window;

	///
	int y_window;

	///
	int w_window;

	///
	int h_window;
	//@}
	
	/// to normalize by mean/stdv
	real *integralimage2;

	
	//-----
	
	/** create a Haar-like feature
	
	    @param width_pattern_ is the #width# of the pattern
	    @param height_pattern_ is the #height# of the pattern
	    @param mask_id_ is the type of the mask
	    @param x_mask_ is the upper left #x# coordinate of the mask
	    @param y_mask_ is the upper left #y# coordinate of the mask
	    @param w_mask_ is the #width# of the mask
	    @param h_mask_ is the #height# of the mask
	*/
	ipHaar(int width_pattern_, int height_pattern_, int mask_id_, int x_mask_, int y_mask_, int w_mask_, int h_mask_);

	///
	virtual ~ipHaar();

	///
	void resetImage(int width_image_, int height_image_);

	///
	void resetWindowX(int x_window_);

	///
	void resetWindowY(int y_window_);

	///
	void resetWindowXY(int x_window_, int y_window_);

	///
	void resetWindowSize(int w_window_, int h_window_);

	///
	virtual void process(Sequence *seq_in);

	/// compute the feature
	virtual real applyMask(Sequence *seq_in) = 0;
	
	/** @name support points of Haar-like features */
	//@
	///
	real p1(Sequence *seq_in);
	///
	real p2(Sequence *seq_in);
	///
	real p3(Sequence *seq_in);
	///
	real p4(Sequence *seq_in);
	///
	real p5(Sequence *seq_in);
	///
	real p6(Sequence *seq_in);
	///
	real p7(Sequence *seq_in);
	///
	real p8(Sequence *seq_in);
	///
	real p9(Sequence *seq_in);
	///
	real p12(Sequence *seq_in);
	///
	real p23(Sequence *seq_in);
	///
	real p14(Sequence *seq_in);
	///
	real p36(Sequence *seq_in);
	///
	real p47(Sequence *seq_in);
	///
	real p69(Sequence *seq_in);
	///
	real p78(Sequence *seq_in);
	///
	real p89(Sequence *seq_in);
	///
	real p15(Sequence *seq_in);
	///
	real p35(Sequence *seq_in);
	///
	real p75(Sequence *seq_in);
	///
	real p95(Sequence *seq_in);
	//@}
	
	///
	virtual void loadXFile(XFile *file);

	///
        virtual void saveXFile(XFile *file);
   
	///
	void info();
};

/** This class is designed to compute a bi-vertical Haar-like feature

	feature = 2.P8 + P1 - 2.P2 - P7 - P9 + P3

    @author Yann Rodriguez (yann.rodriguez@idiap.ch)
    @author Sebastien Marcel (marcel@idiap.ch)
    @version 2.0
    \Date
    @since 2.0
*/
class biVertHaar : public ipHaar
{
public:
   	///
	biVertHaar(int width_pattern_, int height_pattern_, int x_, int y_, int w_, int h_);

	///
	virtual real applyMask(Sequence *seq_in);

	///
	virtual ~biVertHaar() {};
};

/** This class is designed to compute a bi-horizontal Haar-like feature

	feature = 2.P6 + P1 - P3 - 2.P4 - P9 + P7

    @author Yann Rodriguez (yann.rodriguez@idiap.ch)
    @author Sebastien Marcel (marcel@idiap.ch)
    @version 2.0
    \Date
    @since 2.0
*/
class biHorizHaar : public ipHaar
{
public:
	///
	biHorizHaar(int width_pattern_, int height_pattern_, int x_, int y_, int w_, int h_);

	///
	virtual real applyMask(Sequence *seq_in);
	
	///
	virtual ~biHorizHaar() {};
};

/** This class is designed to compute a diagonal Haar-like feature

	feature = P1 - 2.P2 + P3 - 2.P4 + 4.P5 - 2.P6 + P7 - 2.P8 + P9

    @author Yann Rodriguez (yann.rodriguez@idiap.ch)
    @author Sebastien Marcel (marcel@idiap.ch)
    @version 2.0
    \Date
    @since 2.0
*/
class diagHaar : public ipHaar
{
public:
   	///
	diagHaar(int width_pattern_, int height_pattern_, int x_, int y_, int w_, int h_);

	///
	virtual real applyMask(Sequence *seq_in);
	
	///
	virtual ~diagHaar() {};
};

/** This class is designed to compute a tri-vertical Haar-like feature

	feature = P9 + P1 - P3 - P7 - 3.(P89 + P12 - P23 - P78)

    @author Yann Rodriguez (yann.rodriguez@idiap.ch)
    @author Sebastien Marcel (marcel@idiap.ch)
    @version 2.0
    \Date
    @since 2.0
*/
class triVertHaar : public ipHaar
{
public:
   	///
	triVertHaar(int width_pattern_, int height_pattern_, int x_, int y_, int w_, int h_);

	///
	virtual real applyMask(Sequence *seq_in);
	
	///
	virtual ~triVertHaar() {};
};

/** This class is designed to compute a tri-horizontal Haar-like feature

	feature = P9 + P1 - P3 - P7 - 3.(P69 + P14 - P36 - P47)

    @author Yann Rodriguez (yann.rodriguez@idiap.ch)
    @author Sebastien Marcel (marcel@idiap.ch)
    @version 2.0
    \Date
    @since 2.0
*/
class triHorizHaar : public ipHaar
{
public:
   	///
	triHorizHaar(int width_pattern_, int height_pattern_, int x_, int y_, int w_, int h_);

	///
	virtual real applyMask(Sequence *seq_in);
	
	///
	virtual ~triHorizHaar() {};
};

/** This class is designed to compute a central Haar-like feature
 
        feature = P9 + P1 - P3 - P7 - 9.(P95 + P15 - P35 - P75)

    @author Yann Rodriguez (yann.rodriguez@idiap.ch)
    @author Sebastien Marcel (marcel@idiap.ch)
    @version 2.0
    \Date
    @since 2.0
*/
class centralHaar : public ipHaar
{
public:
   	///
	centralHaar(int width_pattern_, int height_pattern_, int x_, int y_, int w_, int h_);
	
	///
	virtual real applyMask(Sequence *seq_in);
	
	///
	virtual ~centralHaar() {};
};

}

#endif
