#include "DirectLDATrainer.h"
#include "DiskXFile.h"
#include "vision.h"

namespace Torch {
	
DirectLDATrainer::DirectLDATrainer(int n_classes_, LDAMachine *machine_) : LDATrainer(n_classes_, machine_)
{
}

void DirectLDATrainer::train(DataSet *data_, MeasurerList *measurers_)
{
   	// computes between-scatter matrix Sb and within-scatter matrix Sw
	makeSbSw(data_);

	if(verbose_level >= 1)
	{
	   	print("\n\n");
   		message("DirectLDATrainer::train ...\n");
	}

	//
	// Computing inverse of Sw
	Mat *Sw_inv = NULL;
	Sw_inv = new(allocator) Mat(n_inputs, n_inputs);

	if(!within_is_identity)
	{
		if(verbose_level >= 1)
			message("Computing inverse of Sw ...");
		mxInverse(Sw, Sw_inv);
	
		if(verbose_level >= 3)
			printM(Sw_inv);

		//
		// Computing Sw^-1 * Sb
		if(verbose_level >= 1)
			message("Computing Sw^-1 * Sb ...");
		mxMatMulMat(Sw_inv, Sb, M_);
	
		if(verbose_level >= 3)
			printM(M_);
	}
	
	//
	// Computing eigenvalues and eigenvectors
	if(verbose_level >= 1)
		message("Computing eigenvalues and eigenvectors ...");

	if(within_is_identity) mxSymEig(Sb, eigenvectors, eigenvalues);
	else mxSymEig(M_, eigenvectors, eigenvalues);

	if(verbose_level >= 3)
	{
		printV(eigenvalues);
		printM(eigenvectors);
	}

	//
	// Indexing eigenvalues and eigenvectors
	if(verbose_level >= 1)
		message("Indexing eigenvalues and eigenvectors...");

	int_double *index_eigenvalues = (int_double*)allocator->alloc(n_inputs*sizeof(int_double));

	index_eigenvalues[0].the_real = eigenvalues->ptr[0];
	index_eigenvalues[0].the_int = 0;
	
	bool need_to_sort = false;
	
	for(int i = 1 ; i < n_inputs ; i++)
	{
		index_eigenvalues[i].the_real = eigenvalues->ptr[i];
		index_eigenvalues[i].the_int = i;
		
		if(index_eigenvalues[i].the_real > index_eigenvalues[i-1].the_real) need_to_sort = true;
		
		if(verbose_level >= 3)
			printf(" -> %g [%d]\n", index_eigenvalues[i].the_real, index_eigenvalues[i].the_int);
	}

	if(need_to_sort)
	{
		//
		// Sort eigenvalues
		if(verbose_level >= 1)
			message("Sorting eigenvalues ...");

		qsort(index_eigenvalues, n_inputs, sizeof(int_double), sort_dec_int_double);
	}
	
	if(verbose_level >= 3)
		for(int i = 0 ; i < n_inputs ; i++)
			printf(" %g [%d]\n", index_eigenvalues[i].the_real, index_eigenvalues[i].the_int);
	
	//
	for(int i = 0 ; i < n_inputs ; i++)
	{
	   	int index_ = index_eigenvalues[i].the_int;

		if((verbose_level >= 3) && (need_to_sort))
			printf(" Swap %d <- %d\n", i, index_eigenvalues[i].the_int);
		
		//
		eigenvalues_[i] = eigenvalues->ptr[index_];
		
		//
		double z, sum = 0.0;
	
		// copy column vector #index_# of eigenvectors to column vector #i# of eigenvectors_
       		for(int j = 0 ; j < n_inputs ; j++)
		{
		   	z = eigenvectors->ptr[j][index_];

			eigenvectors_[j][i] = z;

			sum += z * z;
		}

		// computes norm of the column vector
		double len = sqrt(sum);

		// if norm is not 1 then normalize the column vector
		if(!IS_NEAR(len, 1., 0.01)) 
		{
			warning("Eigenvector is not unit (lenght=%g).", len);
		
			//
       			for(int j = 0 ; j < n_inputs ; j++)
				eigenvectors_[j][i] = eigenvectors->ptr[j][index_] / len;
		}
	}

	//
	// Computes projection of mean into lda-space
	if(verbose_level >= 1)
		message("Computes projection of mean into lda-space ...");
	
	Vec *vinput = new(allocator) Vec(n_inputs);

	for(int i = 0 ; i < n_inputs ; i++)
	{
		vinput->ptr[i] = mean_class_mean_[i];
		d_m_mean_[i] = 0.0;
	}

	if(verbose_level >= 3)
	{
		printV(n_inputs, mean_class_mean_);
		printV(vinput);
	}

	// Y = eigenvectors^T * X
	for(int i = 0 ; i < n_inputs ; i++)
	{
       		for(int j = 0 ; j < n_inputs ; j++)
           		d_m_mean_[i] += eigenvectors->ptr[j][i] * vinput->ptr[j];
	}
	
	if(verbose_level >= 3)
		printV(n_inputs, d_m_mean_);
	
	//
	// Free
	allocator->free(vinput);
	allocator->free(Sw_inv);
}

DirectLDATrainer::~DirectLDATrainer()
{
}

}
