#include "LDATrainer.h"
#include "DiskXFile.h"

namespace Torch {
	
LDATrainer::LDATrainer(int n_classes_, LDAMachine *machine_) : Trainer(machine_)
{
	//
	n_inputs = machine_->n_inputs;
	n_classes = n_classes_;

	//
	mean_ = (real **) allocator->alloc(sizeof(real*) * n_classes);
	for(int i = 0 ; i < n_classes ; i++)
		mean_[i] = (real *) allocator->alloc(sizeof(real) * n_inputs);

	//
	mean_class_mean_ = NULL;
	mean_class_mean_ = (real *) allocator->alloc(sizeof(real) * n_inputs);

	//
    	Sb = NULL;
   	Sb = new(allocator) Mat(n_inputs, n_inputs);

	//
    	Sw = NULL;
   	Sw = new(allocator) Mat(n_inputs, n_inputs);

	M_ = NULL;
	M_ = new(allocator) Mat(n_inputs, n_inputs);

	//
        eigenvalues = NULL;
        eigenvalues = new(allocator) Vec(n_inputs);

        //
        eigenvectors = NULL;
        eigenvectors = new(allocator) Mat(n_inputs, n_inputs);

	//
	// From the LDAMachine
	d_m_mean_ = machine_->d_m_mean_;
    	eigenvalues_ = machine_->eigenvalues;
    	eigenvectors_ = machine_->eigenvectors;
	
	//
	addIOption("verbose_level", &verbose_level, 0, "verbose level");
	addBOption("class provided", &classprovided, false, "class are provided");
	addBOption("within identity", &within_is_identity, false, "within_is_identity");
	addBOption("save", &save, false, "save Sb and Sw matrix");
}

void LDATrainer::makeSbSw(DataSet *data_)
{
   	int n_data = data_->n_examples;

	if(verbose_level >= 1)
	{
	   	print("\n\n");
   		message("LDATrainer::makeSbSw() ...\n");
	}
	
	if(verbose_level >= 2)
	{
	   	print("n_inputs = %d\n", data_->n_inputs);
   		print("n_examples = %d\n", n_data);
	}

	if(n_inputs != data_->n_inputs) error("LDATrainer::makeSbSw() incorrect number of inputs (%d != %d)\n", n_inputs, data_->n_inputs);


	
	//
	// Init

	int *label = NULL;
	int *n_samples = NULL;
	int n_samples_per_class = 0;

	if(classprovided)
	{
		label = (int *) allocator->alloc(sizeof(int) * n_classes);
		n_samples = (int *) allocator->alloc(sizeof(int) * n_classes);
	}
	else
	{
		n_samples_per_class = n_data / n_classes;

		if(n_samples_per_class * n_classes != n_data)
			error("LDATrainer::makeSbSw() incorrect number of examples per classes (%d*%d != %d)", n_samples_per_class, n_classes, n_data);

		print("There is %d examples per classes\n", n_samples_per_class);
	}

	for(int c = 0 ; c < n_classes ; c++)
	{
		if(classprovided)
		{
	   		label[c] = 0;
	   		n_samples[c] = 0;
		}

		for(int row = 0 ; row < n_inputs ; row++) mean_[c][row] = 0.0;
	}


	
	//
	// Init grand mean
	for(int row = 0 ; row < n_inputs ; row++) mean_class_mean_[row] = 0.0;

	//
	// Init means of class
	for(int col = 0 ; col < n_data ; col++)
	{
	   	// set current example
	   	data_->setExample(col);
		
		// get the class
		int c;
		if(classprovided) c = (int) data_->targets->frames[0][0];
		else c = col / n_samples_per_class;

		// check the class
		if(c < 0) error("Incorrect target class.");
		if(c >= n_classes) error("Incorrect target class.");
		
		// copy the example into the #data# matrix
		for(int row = 0 ; row < n_inputs ; row++)
		{
		   	real z = data_->inputs->frames[0][row];
			
			mean_[c][row] += z;
			mean_class_mean_[row] += z;
		}

		if(classprovided)
		{
			label[c] = 1;
			n_samples[c]++;
		}
	}
	
	if(classprovided)
	{
		for(int c = 0 ; c < n_classes ; c++)
		{
			if(label[c] == 0) error("Expected target class missing.");

			if(verbose_level >= 2)
				print("Class %d: %d samples\n", c, n_samples[c]);
		}
	}

	
	//
	// Computing means of class
	if(verbose_level >= 1)
		message("Computing means ...");
	
	for(int c = 0 ; c < n_classes ; c++)
	{
		float n_samples_;

		if(classprovided) n_samples_ = (float) n_samples[c];
		else n_samples_ = (float) n_samples_per_class;

		for(int row = 0 ; row < n_inputs ; row++)
		   	mean_[c][row] /= n_samples_;
	
		if(verbose_level >= 3)
			printV(n_inputs, mean_[c]);
	}

	for(int row = 0 ; row < n_inputs ; row++) mean_class_mean_[row] /= (float) n_data;

	if(verbose_level >= 3)
		printV(n_inputs, mean_class_mean_);
	
	//
	// Computing the Sb covariance matrix
	if(verbose_level >= 1)
		message("Computing the between-class covariance matrix ...");

	Vec *V_ = NULL;
	V_ = new(allocator) Vec(n_inputs);
	
   	Sb->zero();

	for(int c = 0 ; c < n_classes ; c++)
	{
	   	// V_ = mean[i] - mean_class_mean_ (\mu_i - \mu)
		for(int row = 0 ; row < n_inputs ; row++)
	   		V_->ptr[row] = mean_[c][row] - mean_class_mean_[row];
		
		M_->zero();
		
		// M_ = V_ * V_^T
		mxVecMulTrVec(V_, V_, M_);

		real prior = (real) n_samples[c] / (real) n_classes;
		
		//
		mxRealMulMat(prior, M_, M_);
		
		// Madd += M_
		mxMatAddMat(Sb, M_, Sb);
	}

	if(verbose_level >= 3)
		printM(Sb);
   	
	if(save)
	{
		message("Saving Sb matrix...");

   		DiskXFile *file = NULL;

		file = new DiskXFile("Sb.mat", "w");
	
		file->printf("%d %d\n", Sb->m, Sb->n);
		
		for(int m_ = 0 ; m_ < Sb->m ; m_++)
		{
			for(int n_ = 0 ; n_ < Sb->n ; n_++)
				file->printf("%2.9f ", Sb->ptr[m_][n_]);
			file->printf("\n");
		}

		delete file;
	}
	

	if(!within_is_identity)
	{  
	//
	// Computing the Sw covariance matrix
	if(verbose_level >= 1)
		message("Computing the within-class covariance matrix ...");
	
	Sw->zero();
	
	for(int col = 0 ; col < n_data ; col++)
	{
	   	// set current example
	   	data_->setExample(col);
		
		// get the class
		int c;
		if(classprovided) c = (int) data_->targets->frames[0][0];
		else c = col / n_samples_per_class;

	   	// V_ = mean[i] - mean_class_mean_ (\mu_i - \mu)
		for(int row = 0 ; row < n_inputs ; row++)
	   		V_->ptr[row] = data_->inputs->frames[0][row] - mean_[c][row];

		M_->zero();
		
		// M_ = V_ * V_^T
		mxVecMulTrVec(V_, V_, M_);

		// Madd += M_
		mxMatAddMat(Sw, M_, Sw);
	}
	
	mxRealMulMat(1.0/(float) n_data, Sw, Sw);
	
	if(verbose_level >= 3)
		printM(Sw);

	if(save)
	{
		message("Saving Sw matrix...");

   		DiskXFile *file = NULL;

		file = new DiskXFile("Sw.mat", "w");
	
		file->printf("%d %d\n", Sw->m, Sw->n);
		
		for(int m_ = 0 ; m_ < Sw->m ; m_++)
		{
			for(int n_ = 0 ; n_ < Sw->n ; n_++)
				file->printf("%2.9f ", Sw->ptr[m_][n_]);
			file->printf("\n");
		}

		delete file;
	}
	
	}
	
	//
	allocator->free(V_);
	if(classprovided)
	{
		allocator->free(n_samples);
		allocator->free(label);
	}
}

LDATrainer::~LDATrainer()
{
}

}
