#include "PCATrainer.h"
#include "DiskXFile.h"
#include "vision.h"

namespace Torch {

PCATrainer::PCATrainer(PCAMachine *machine_) : Trainer(machine_)
{
	//
	n_inputs = machine_->n_inputs;
	
	//
	CovMat = NULL;
	CovMat = new(allocator) Mat(n_inputs, n_inputs);

	//
        eigenvalues = NULL;
        eigenvalues = new(allocator) Vec(n_inputs);

        //
        eigenvectors = NULL;
        eigenvectors = new(allocator) Mat(n_inputs, n_inputs);

	//
	// From the PCAMachine
   	Xm_ = machine_->Xm;
    	eigenvectors_ = machine_->eigenvectors;
    	eigenvalues_ = machine_->eigenvalues;
	
	//
	addIOption("verbose_level", &verbose_level, 0, "verbose level");
	addBOption("saveCovar", &saveCovar, false, "save the covariance matrix");
}

void PCATrainer::train(DataSet *data_, MeasurerList *measurers_)
{
   	int n_data = data_->n_examples;

	if(verbose_level >= 1)
   		message("PCATrainer::train ...\n");

	if(verbose_level >= 2)
	{
   		print("n_inputs = %d\n", data_->n_inputs);
	   	print("n_examples = %d\n", n_data);
	}

	if(n_inputs != data_->n_inputs) error("PCATrainer::train() incorrect number of inputs (%d != %d)\n", n_inputs, data_->n_inputs);
	
	if(n_data < n_inputs)
	{
		warning("n_examples (%d) < n_inputs (%d), you should use more data to train the PCA.", n_data, n_inputs);

		/* in this case, we should use the following method:

			let X be the zero mean data

			1. computes eigenvectors E' and eigenvalues V of X^T X
			2. eigenvectors E are given by E = X E' V^{-1}
		*/
	}
	
	//
	// Computes the mean of the training data
	if(verbose_level >= 1)
		message("Computes the mean the training data ...");

	//
      	for(int row = 0 ; row < n_inputs ; row++) Xm_[row] = 0.0;
	   
   	for(int col = 0 ; col < n_data ; col++) 
	{
   		data_->setExample(col);
		
      		for(int row = 0 ; row < n_inputs ; row++)
			Xm_[row] += data_->inputs->frames[0][row];
	}

      	for(int row = 0 ; row < n_inputs ; row++) Xm_[row] /= (float) n_data;
	
	
	//
	// Computing the covariance matrix
	if(verbose_level >= 1)
		message("Computing the covariance matrix ...");
   
	Vec *V_ = NULL;
	V_ = new(allocator) Vec(n_inputs);
	Mat *M_ = NULL;
	M_ = new(allocator) Mat(n_inputs, n_inputs);

	CovMat->zero();

	for(int col = 0 ; col < n_data ; col++)
	{
	   	data_->setExample(col);
		
		for(int row = 0 ; row < n_inputs ; row++)
			V_->ptr[row] = data_->inputs->frames[0][row] - Xm_[row];

		M_->zero();
		
		// M_ = V_ * V_^T
		mxVecMulTrVec(V_, V_, M_);

		// Madd += M_
		mxMatAddMat(CovMat, M_, CovMat);
	}

	mxRealMulMat(1.0/(float) n_data, CovMat, CovMat);
	
	if(verbose_level >= 3)
		printM(CovMat);

	//
	allocator->free(V_);
	allocator->free(M_);

	if(saveCovar)
	{
		message("Saving Covariance matrix...");

   		DiskXFile *file = NULL;

		file = new DiskXFile("covar", "w");
		
		for(int m_ = 0 ; m_ < CovMat->m ; m_++)
			for(int n_ = 0 ; n_ < CovMat->n ; n_++)
				file->write(&CovMat->ptr[m_][n_], sizeof(double), 1);

		delete file;
	}
	
	//
	// Computing eigenvalues and eigenvectors
	if(verbose_level >= 1)
		message("Computing eigenvalues and eigenvectors...");
   	mxSymEig(CovMat, eigenvectors, eigenvalues);
 
	if(verbose_level >= 3)
	{
		printV(eigenvalues);
		printM(eigenvectors);
	}
	
	//
	// Indexing eigenvalues and eigenvectors
	if(verbose_level >= 1)
		message("Indexing eigenvalues and eigenvectors...");

	int_double *index_eigenvalues = (int_double*)allocator->alloc(n_inputs*sizeof(int_double));

	index_eigenvalues[0].the_real = eigenvalues->ptr[0];
	index_eigenvalues[0].the_int = 0;
	
	bool need_to_sort = false;
	
	for(int i = 1 ; i < n_inputs ; i++)
	{
		index_eigenvalues[i].the_real = eigenvalues->ptr[i];
		index_eigenvalues[i].the_int = i;
		
		if(index_eigenvalues[i].the_real > index_eigenvalues[i-1].the_real) need_to_sort = true;
		
		if(verbose_level >= 3)
			printf(" -> %g [%d]\n", index_eigenvalues[i].the_real, index_eigenvalues[i].the_int);
	}

	if(need_to_sort)
	{
		//
		// Sort eigenvalues
		if(verbose_level >= 1)
			message("Sorting eigenvalues ...");

		qsort(index_eigenvalues, n_inputs, sizeof(int_double), sort_dec_int_double);
	}
	
	if(verbose_level >= 3)
		for(int i = 0 ; i < n_inputs ; i++)
			printf(" %g [%d]\n", index_eigenvalues[i].the_real, index_eigenvalues[i].the_int);

	for(int i = 0 ; i < n_inputs ; i++)
	{
	   	int index_ = index_eigenvalues[i].the_int;
	
		if((verbose_level >= 3) && (need_to_sort))
			printf(" Swap %d <- %d\n", i, index_eigenvalues[i].the_int);
		
		//
		eigenvalues_[i] = eigenvalues->ptr[index_];
		
		//
		double z, sum = 0.0;
		
       		for(int j = 0 ; j < n_inputs ; j++)
		{
		   	z = eigenvectors->ptr[j][index_];

			eigenvectors_[j][i] = z;

			sum += z * z;
		}

		double len = sqrt(sum);

		if(!IS_NEAR(len, 1.0, 0.01))
		{
			warning("Eigenvector is not unit (lenght=%g).", len);

       			//
			for(int j = 0 ; j < n_inputs ; j++)
		   		eigenvectors_[j][i] = eigenvectors->ptr[j][index_] / len;
		}
	}

	allocator->free(index_eigenvalues);
}

PCATrainer::~PCATrainer()
{
}

}
