#include "StumpTrainer.h"

namespace Torch {

StumpTrainer::StumpTrainer(StumpMachine *stump_machine_) : Trainer(stump_machine_)
{
	addBOption("verbose", &verbose, false, "verbose");

   	stump_machine = stump_machine_;

	bestFeature = 0;
	bestThreshold = 0.;
	bestDirection = 0;

	n_pos = 0;
	n_neg = 0;
    
	shift = 0;
	sample = 0;
	min_err = INF;
}

void StumpTrainer::train(DataSet *data_, MeasurerList *measurers)  
{	
	bestFeature = 0;
	bestThreshold = 0.;
	bestDirection = 0;
  	min_err = INF;
	
	int n_examples = data_->n_examples;

	features = (real *)allocator->alloc(sizeof(real)*n_examples);
	targets = (real *)allocator->alloc(sizeof(real)*n_examples);

	for(int feat = 0; feat < stump_machine->n_features; feat++)
	{
		for(int j = 0 ; j < HIST_PRECISION_; j++)
		{
			histPos[j] = 0.;
			histNeg[j] = 0.;
			hCumulPos[j] = 0.;
			hCumulNeg[j] = 0.;
		}
		
		computeHistogram(data_, feat);
		computeBestThreshold(feat);
	}

	if(verbose)
		message("\n StumpTrainer:  feature -> %d  threshold -> %2.2f  direction %d", bestFeature, bestThreshold, bestDirection);

	stump_machine->setParams(bestFeature, bestThreshold, bestDirection);

	allocator->free(features);
	allocator->free(targets);
}

void StumpTrainer::computeHistogram(DataSet *data_, int feature_id)
{
	int n_examples = data_->n_examples;

	//
	real min = INF, max = -INF;

	for(int t = 0; t < n_examples; t++)
	{
		data_->setExample(t);

		features[t] = stump_machine->getFeature(feature_id, data_->inputs);
		targets[t] = data_->targets->frames[0][0];
		
		if(features[t]<min) min = features[t];
		else if(features[t]>max) max = features[t];
	}

	sample = (max - min)/HIST_PRECISION_;
	shift = -min;

	//
	n_pos = 0;
	n_neg = 0;
	real value_, label;

	for(int j=0; j<n_examples; j++)
	{
		label = targets[j];
		value_ = features[j] + shift;

		if(label == 1)  /* it is a positive pattern */
		{
			n_pos++;

			int index = (int)(floor(value_/sample));
			
			if(index < 0) index = 0;
			if(index >= HIST_PRECISION_) index = HIST_PRECISION_-1;	

			histPos[index]++;
		}
		else if(label == -1)  /* it's a negative pattern */
		{
			n_neg++;

			int index = (int)(floor(value_/sample));

			if(index < 0) index = 0;
			if(index >= HIST_PRECISION_) index = HIST_PRECISION_-1;	

			histNeg[index]++;
		}
	}
}
  
void StumpTrainer::computeBestThreshold(int i)
{   
   	/*
	  Here we need to find the threshold that yields the lowest error.
	  The error is obtained directly from the cumulative histograms.
      
	  For each threshold, compute the error we get with h+ and with h- and keep the best of the two,
	  then keep the threshold with the least error.
	*/

	float error1;
	float error2;
	float error;
	int direction;
   	
	real inv_n_pos = 1./(n_pos);
	real inv_n_neg = 1./(n_neg);

	real hCumulPos_ = 0.0;
	real hCumulNeg_ = 0.0;

	for(int t = 0; t < HIST_PRECISION_; t++)
	{
		// First we have to normalize the histograms with respect to the proportion of positive and negative patterns
		hCumulPos_ += histPos[t] * inv_n_pos;
		hCumulNeg_ += histNeg[t] * inv_n_neg;

		// compute cumulative histograms
		hCumulPos[t] = hCumulPos_;
		hCumulNeg[t] = hCumulNeg_;

		error1 = 0;  // assuming that the positive patterns are > threshold
		error2 = 0;  // assuming that the negative patterns are < threshold
		error = 0;
		direction = 0;
      
		error1 = ((1. - hCumulNeg[t]) * n_neg) + (hCumulPos[t] * n_pos);
		error2 = (hCumulNeg[t] * n_neg) + ((1. - hCumulPos[t]) * n_pos);
		
		if(error1 <= error2)
		{
			error = error1;
			direction = 1;
		}
		else
		{
			error = error2;
			direction = -1;
		}
		if(error < min_err)
		{
			bestThreshold = (t+1)*sample-shift;
			bestDirection = direction;
			bestFeature = i;
			min_err = error;
		}
	}
}	

StumpTrainer::~StumpTrainer()
{
}

}
